# Generated by Django 2.2.6 on 2019-10-03 21:12

import uuid

from django.conf import settings
from django.db import migrations, models
import django.contrib.postgres.fields.ranges
import django.db.models.deletion
import oscar.models.fields

import oscarreports.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Report",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        unique=True,
                        verbose_name="UUID",
                    ),
                ),
                (
                    "type_code",
                    models.CharField(max_length=50, verbose_name="Type Code"),
                ),
                ("description", models.TextField(verbose_name="Description")),
                (
                    "date_range",
                    django.contrib.postgres.fields.ranges.DateTimeRangeField(
                        blank=True, null=True, verbose_name="Date Range"
                    ),
                ),
                (
                    "task_id",
                    models.UUIDField(
                        blank=True,
                        editable=False,
                        null=True,
                        unique=True,
                        verbose_name="Background Task ID",
                    ),
                ),
                (
                    "created_on",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created On"),
                ),
                (
                    "queued_on",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Queued On"
                    ),
                ),
                (
                    "started_on",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Started On"
                    ),
                ),
                (
                    "completed_on",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Completed On"
                    ),
                ),
                (
                    "mime_type",
                    oscar.models.fields.NullCharField(
                        max_length=20, verbose_name="MIME Type"
                    ),
                ),
                (
                    "report_file",
                    models.FileField(
                        blank=True,
                        null=True,
                        upload_to=oscarreports.models.get_report_upload_path,
                        verbose_name="Report File",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.ContentType",
                        verbose_name="Content Type",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Owner",
                    ),
                ),
            ],
        ),
    ]



from django.urls import path
from django.urls.resolvers import URLPattern, URLResolver
from oscar.apps.dashboard.reports import apps


class ReportsDashboardConfig(apps.ReportsDashboardConfig):
    name = "oscarreports"

    def ready(self) -> None:
        super().ready()
        from . import handlers, views  # NOQA

        self.index_view = views.IndexView
        self.download_view = views.ReportDownloadView
        self.delete_view = views.ReportDeleteView

    def get_urls(self) -> list[URLPattern | URLResolver]:
        urls: list[URLPattern | URLResolver] = [
            path("", self.index_view.as_view(), name="reports-index"),
            path(
                "<uuid:uuid>/download/",
                self.download_view.as_view(),
                name="reports-download",
            ),
            path(
                "<uuid:uuid>/delete/",
                self.delete_view.as_view(),
                name="reports-delete",
            ),
        ]
        return self.post_process_urls(urls)



from datetime import timedelta

from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.utils import timezone
from freezegun import freeze_time

from .. import models, tasks

try:
    try:
        from psycopg.types.range import Range as DateTimeTZRange
    except ImportError:
        from psycopg2.extras import DateTimeTZRange
except ImportError:
    raise ImproperlyConfigured("Error loading psycopg2 or psycopg module")


@freeze_time("2019-10-03T12:00:00-04:00")
class ReportTest(TestCase):
    def setUp(self) -> None:
        self.staff_user = User.objects.create_user(username="root", is_staff=True)

        self.report = models.Report()
        self.report.uuid = "d3c74a8b-e7ae-4482-bd9c-bee69fde5c5c"
        self.report.content_type = None
        self.report.owner = self.staff_user
        self.report.type_code = "order_report"
        self.report.date_range = DateTimeTZRange(
            lower=(timezone.now() - timedelta(days=1)), upper=(timezone.now())
        )
        self.report.save()

    def test_generate_report(self) -> None:
        self.assertIsNone(self.report.started_on)
        self.assertIsNone(self.report.mime_type)
        self.assertIsNone(self.report.report_file.name)
        self.assertIsNone(self.report.completed_on)

        tasks.generate_report.enqueue(str(self.report.uuid), "CSV")

        self.report.refresh_from_db()

        self.assertIsNotNone(self.report.started_on)
        self.assertEqual(self.report.mime_type, "text/csv")
        self.assertTrue(
            self.report.report_file.name.startswith(
                "oscar-reports/2019/10/03/d3c74a8b-e7ae-4482-bd9c-bee69fde5c5c"
            )
        )
        self.assertTrue(self.report.report_file.name.endswith(".csv"))
        self.assertIsNotNone(self.report.completed_on)



from __future__ import annotations

from typing import TYPE_CHECKING
import io
import os.path
import uuid

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import DateTimeRangeField
from django.contrib.sites.models import Site
from django.core.mail import EmailMultiAlternatives
from django.db import models, transaction
from django.http import HttpResponse
from django.template.loader import get_template
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_stubs_ext import StrOrPromise
from oscar.apps.dashboard.reports.reports import ReportGenerator
from oscar.models.fields import NullCharField

from . import tasks

if TYPE_CHECKING:
    from django_tasks import TaskResult

from .utils import GeneratorRepository


def get_report_upload_path(instance: Report, filename: str) -> str:
    # Upload files to {MEDIA_ROOT}/{OSCAR_REPORTS_UPLOAD_PREFIX}/{YYYY}/{MM}/{DD}/{uuid}.{ext}
    prefix = getattr(settings, "OSCAR_REPORTS_UPLOAD_PREFIX", "oscar-reports")
    extension = os.path.splitext(filename)[1].replace(".", "")
    return "{prefix}/{date}/{uuid}.{ext}".format(
        prefix=prefix,
        date=instance.created_on.strftime("%Y/%m/%d"),
        uuid=instance.uuid,
        ext=extension,
    )


class Report(models.Model):
    STATUS_CREATED = "created"
    STATUS_QUEUED = "queued"
    STATUS_IN_PROGRESS = "in-progress"
    STATUS_COMPLETED = "completed"
    STATUS_NAMES = {
        STATUS_CREATED: _("Created"),
        STATUS_QUEUED: _("Queued"),
        STATUS_IN_PROGRESS: _("In-Progress"),
        STATUS_COMPLETED: _("Completed"),
    }

    # Unique ID used in URLs and filenames
    uuid = models.UUIDField(_("UUID"), default=uuid.uuid4, editable=False, unique=True)

    # Store a link to the content-type (model) which was used to generate this report.
    # Only user's with the `can_view` permission on this content-type will be able to
    # see this report.
    content_type = models.ForeignKey(
        ContentType,
        verbose_name=_("Content Type"),
        on_delete=models.CASCADE,
        related_name="+",
        null=True,
        blank=True,
    )

    # Report Metadata
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Owner"),
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    type_code = models.CharField(_("Type Code"), max_length=50)
    description = models.TextField(_("Description"))
    date_range = DateTimeRangeField(  # type:ignore[misc]
        _("Date Range"),
        null=True,
        blank=True,
    )

    # Background Task ID
    task_id = models.CharField(
        _("Background Task ID"),
        editable=False,
        unique=True,
        null=True,
        blank=True,
        max_length=64,
    )

    # Status Timestamps
    created_on = models.DateTimeField(_("Created On"), auto_now_add=True)
    queued_on = models.DateTimeField(_("Queued On"), null=True, blank=True)
    started_on = models.DateTimeField(_("Started On"), null=True, blank=True)
    completed_on = models.DateTimeField(_("Completed On"), null=True, blank=True)

    # Report File Output
    mime_type = NullCharField(_("MIME Type"), max_length=20)
    report_file = models.FileField(
        _("Report File"), upload_to=get_report_upload_path, null=True, blank=True
    )

    def __str__(self) -> str:
        return str(self.uuid)

    @property
    def status(self) -> str:
        if self.completed_on:
            return self.STATUS_COMPLETED
        if self.started_on:
            return self.STATUS_IN_PROGRESS
        if self.queued_on:
            return self.STATUS_QUEUED
        return self.STATUS_CREATED

    @property
    def status_name(self) -> StrOrPromise:
        return self.STATUS_NAMES.get(self.status, "")

    @property
    def is_complete(self) -> bool:
        return self.status == self.STATUS_COMPLETED

    @property
    def generator_class(
        self,
    ) -> type[ReportGenerator]:
        generator = GeneratorRepository().get_generator(self.type_code)
        if not generator:
            raise ValueError("Can not queue report: no generator class was found.")
        return generator

    @property
    def task_result(self) -> TaskResult[None] | None:
        if not self.task_id:
            return None
        return tasks.generate_report.get_result(self.task_id)

    @property
    def task_status(self) -> str | None:
        result = self.task_result
        if result is None:
            return None
        return result.status.name.replace("_", " ").title()

    def queue(
        self,
        report_format: str = "CSV",
    ) -> None:
        # Reset metadata
        generator = self.get_generator(report_format)
        self.description = generator.report_description()
        self.queued_on = None
        self.started_on = None
        self.completed_on = None
        self.task_id = None
        self.save(
            update_fields=[
                "description",
                "queued_on",
                "started_on",
                "completed_on",
                "task_id",
            ]
        )

        # Defer enqueue AND task_id save until after commit
        def do_enqueue() -> None:
            result = tasks.generate_report.enqueue(str(self.uuid), report_format)
            self.queued_on = timezone.now()
            self.task_id = result.id
            self.save(update_fields=["queued_on", "task_id"])

        transaction.on_commit(do_enqueue)

    queue.alters_data = True  # type:ignore[attr-defined]

    def generate(self, report_format: str = "CSV") -> None:
        # Record start time
        self.started_on = timezone.now()
        self.save(update_fields=["started_on"])
        # Generate report content
        generator = self.get_generator(report_format)
        report = generator.generate()
        if not isinstance(report, HttpResponse):
            raise TypeError(
                f"Expected HttpResponse from generator, got {type(report).__name__}"
            )
        # Save generated content
        filename = self.get_filename(report_format)
        content = io.BytesIO(report.content)
        self.mime_type = report["content-type"]
        self.report_file.save(filename, content)
        self.completed_on = timezone.now()
        self.save(
            update_fields=[
                "mime_type",
                "report_file",
                "completed_on",
            ]
        )
        self.send_completed_alert()

    generate.alters_data = True  # type:ignore[attr-defined]

    def send_completed_alert(self) -> None:
        if self.owner is None or not self.owner.email:
            return
        ctx = {
            "site": Site.objects.get_current(),
            "report": self,
        }
        subject = get_template(
            "oscar/dashboard/reports/emails/report_completed_alert_subject.txt"
        ).render(ctx)
        subject = subject.replace("\n", "")
        text = get_template(
            "oscar/dashboard/reports/emails/report_completed_alert_body.txt"
        ).render(ctx)
        html = get_template(
            "oscar/dashboard/reports/emails/report_completed_alert_body.html"
        ).render(ctx)
        to_addrs = [self.owner.email]
        msg = EmailMultiAlternatives(subject, text, settings.OSCAR_FROM_EMAIL, to_addrs)
        msg.attach_alternative(html, "text/html")
        msg.send()

    send_completed_alert.alters_data = True  # type:ignore[attr-defined]

    def get_generator(
        self,
        report_format: str,
    ) -> ReportGenerator:
        kwargs = {
            "start_date": self.date_range.lower if self.date_range else None,
            "end_date": self.date_range.upper if self.date_range else None,
            "formatter": report_format,
        }
        return self.generator_class(**kwargs)

    def get_filename(self, report_format: str) -> str:
        return f"{self.uuid}.{report_format.lower()}"



from __future__ import annotations

from typing import Any

from django.db.models.signals import pre_delete
from django.dispatch import receiver

from .models import Report


@receiver(pre_delete, sender=Report)
def auto_delete_report_file(
    sender: type[Report],
    instance: Report,
    **kwargs: Any,
) -> None:
    if instance.report_file:
        instance.report_file.delete(save=False)



from __future__ import annotations

from uuid import UUID

from django_tasks import task


@task()
def generate_report(report_uuid: str, report_format: str) -> None:
    from . import models

    report = models.Report.objects.get(uuid=UUID(report_uuid))
    report.generate(report_format)



import os

from django.utils.translation import gettext_lazy as _
from oscar.defaults import *  # noqa
import django_stubs_ext

django_stubs_ext.monkeypatch()

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DEBUG = True
SECRET_KEY = "li0$-gnv)76g$yf7p@(cg-^_q7j6df5cx$o-gsef5hd68phj!4"
SITE_ID = 1

ROOT_URLCONF = "sandbox.urls"
ALLOWED_HOSTS = ["*"]

USE_TZ = True
TIME_ZONE = "America/New_York"
USE_I18N = True
LANGUAGE_CODE = "en-us"
LANGUAGES = (
    ("en-us", _("English")),
    ("es", _("Spanish")),
)

# Configure JUnit XML output
TEST_RUNNER = "xmlrunner.extra.djangotestrunner.XMLTestRunner"
_tox_env_name = os.environ.get("TOX_ENV_NAME")
if _tox_env_name:
    TEST_OUTPUT_DIR = os.path.join(BASE_DIR, f"../../junit-{_tox_env_name}/")
else:
    TEST_OUTPUT_DIR = os.path.join(BASE_DIR, "../../junit/")

INSTALLED_APPS = [
    # Core Django
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.postgres",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.flatpages",
    # Oscar Reports
    "oscarreports",
    # django-oscar
    "oscar.config.Shop",
    "oscar.apps.analytics.apps.AnalyticsConfig",
    "oscar.apps.checkout.apps.CheckoutConfig",
    "oscar.apps.address.apps.AddressConfig",
    "oscar.apps.shipping.apps.ShippingConfig",
    "oscar.apps.catalogue.apps.CatalogueConfig",
    "oscar.apps.catalogue.reviews.apps.CatalogueReviewsConfig",
    "oscar.apps.communication.apps.CommunicationConfig",
    "oscar.apps.partner.apps.PartnerConfig",
    "oscar.apps.basket.apps.BasketConfig",
    "oscar.apps.payment.apps.PaymentConfig",
    "oscar.apps.offer.apps.OfferConfig",
    "oscar.apps.order.apps.OrderConfig",
    "oscar.apps.customer.apps.CustomerConfig",
    "oscar.apps.search.apps.SearchConfig",
    "oscar.apps.voucher.apps.VoucherConfig",
    "oscar.apps.wishlists.apps.WishlistsConfig",
    "oscar.apps.dashboard.apps.DashboardConfig",
    # "oscar.apps.dashboard.reports.apps.ReportsDashboardConfig",
    "oscar.apps.dashboard.users.apps.UsersDashboardConfig",
    "oscar.apps.dashboard.orders.apps.OrdersDashboardConfig",
    "oscar.apps.dashboard.catalogue.apps.CatalogueDashboardConfig",
    "oscar.apps.dashboard.offers.apps.OffersDashboardConfig",
    "oscar.apps.dashboard.partners.apps.PartnersDashboardConfig",
    "oscar.apps.dashboard.pages.apps.PagesDashboardConfig",
    "oscar.apps.dashboard.ranges.apps.RangesDashboardConfig",
    "oscar.apps.dashboard.reviews.apps.ReviewsDashboardConfig",
    "oscar.apps.dashboard.vouchers.apps.VouchersDashboardConfig",
    "oscar.apps.dashboard.communications.apps.CommunicationsDashboardConfig",
    "oscar.apps.dashboard.shipping.apps.ShippingDashboardConfig",
    # 3rd-party apps that oscar depends on
    "widget_tweaks",
    "haystack",
    "treebeard",
    "sorl.thumbnail",
    "django_tables2",
]

INSTALLED_APPS.append("django_tasks")

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "%(asctime)s django %(name)s: %(levelname)s %(process)d %(thread)d %(message)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S",
        }
    },
    "handlers": {"console": {"class": "logging.StreamHandler", "formatter": "verbose"}},
    "loggers": {},
    "root": {
        "handlers": ["console"],
        "level": "ERROR",
    },
}

MIDDLEWARE = (
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "oscar.apps.basket.middleware.BasketMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
                "oscar.apps.search.context_processors.search_form",
                "oscar.apps.checkout.context_processors.checkout",
                "oscar.apps.communication.notifications.context_processors.notifications",
                "oscar.core.context_processors.metadata",
            ],
        },
    },
]

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "postgres",
        "USER": "postgres",
        "PASSWORD": "",
        "HOST": "postgres",
        "PORT": 5432,
    }
}

HAYSTACK_CONNECTIONS = {
    "default": {
        "ENGINE": "haystack.backends.simple_backend.SimpleEngine",
    },
}

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "sandbox",
    }
}

AUTHENTICATION_BACKENDS = (
    "oscar.apps.customer.auth_backends.EmailBackend",
    "django.contrib.auth.backends.ModelBackend",
)


# Static files (CSS, JavaScript, Images)
STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(BASE_DIR, "public", "static")
MEDIA_URL = "/media/"
MEDIA_ROOT = os.path.join(BASE_DIR, "public", "media")

# Order Status Pipeline
# Needed by oscarapicheckout
ORDER_STATUS_PENDING = "Pending"
ORDER_STATUS_PAYMENT_DECLINED = "Payment Declined"
ORDER_STATUS_AUTHORIZED = "Authorized"

# Other statuses
ORDER_STATUS_SHIPPED = "Shipped"
ORDER_STATUS_CANCELED = "Canceled"

OSCAR_INITIAL_ORDER_STATUS = ORDER_STATUS_PENDING
OSCARAPI_INITIAL_ORDER_STATUS = ORDER_STATUS_PENDING
OSCAR_ORDER_STATUS_PIPELINE = {
    ORDER_STATUS_PENDING: (
        ORDER_STATUS_PAYMENT_DECLINED,
        ORDER_STATUS_AUTHORIZED,
        ORDER_STATUS_CANCELED,
    ),
    ORDER_STATUS_PAYMENT_DECLINED: (ORDER_STATUS_CANCELED,),
    ORDER_STATUS_AUTHORIZED: (ORDER_STATUS_SHIPPED, ORDER_STATUS_CANCELED),
    ORDER_STATUS_SHIPPED: (),
    ORDER_STATUS_CANCELED: (),
}

OSCAR_INITIAL_LINE_STATUS = ORDER_STATUS_PENDING
OSCAR_LINE_STATUS_PIPELINE = {
    ORDER_STATUS_PENDING: (ORDER_STATUS_SHIPPED, ORDER_STATUS_CANCELED),
    ORDER_STATUS_SHIPPED: (),
    ORDER_STATUS_CANCELED: (),
}

# Oscar
OSCAR_SHOP_NAME = _("Sandbox")  # type:ignore[assignment]
OSCAR_ALLOW_ANON_CHECKOUT = True
OSCAR_DEFAULT_CURRENCY = "USD"
OSCARAPI_BLOCK_ADMIN_API_ACCESS = False

# Disable real emails
EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"

# Django Tasks Config
TASKS = {
    "default": {
        "BACKEND": "django_tasks.backends.immediate.ImmediateBackend",
        "ENQUEUE_ON_COMMIT": False,
    },
}



from django.utils.translation import gettext_lazy as _
from django_tables2 import Column, TemplateColumn
from oscar.apps.dashboard.tables import DashboardTable


class ReportTable(DashboardTable):
    description = Column(accessor="description", orderable=False)
    date_range = TemplateColumn(
        template_name="oscar/dashboard/reports/report_row_date_range.html",
        verbose_name=_("Date Range"),
        orderable=False,
    )

    owner = Column(
        accessor="owner__get_full_name", orderable=False, verbose_name=_("Owner")
    )
    status = Column(
        accessor="status_name", orderable=False, verbose_name=_("Report Status")
    )
    task_status = Column(
        accessor="task_status", orderable=False, verbose_name=_("Task Status")
    )
    filesize = TemplateColumn(
        template_name="oscar/dashboard/reports/report_row_filesize.html",
        verbose_name=_("File Size"),
        orderable=False,
    )

    queued_on = Column(accessor="queued_on")
    started_on = Column(accessor="started_on")
    completed_on = Column(accessor="completed_on")

    actions = TemplateColumn(
        template_name="oscar/dashboard/reports/report_row_actions.html",
        verbose_name=" ",
        orderable=False,
    )

    class Meta(DashboardTable.Meta):
        pass



from django.test import TestCase

from .. import forms


class ReportFormTest(TestCase):
    def test_date_range_empty(self) -> None:
        form = forms.ReportForm(data={})
        self.assertFalse(form.is_valid())

    def test_date_range_complete(self) -> None:
        form = forms.ReportForm(
            data={
                "date_from": "2016-11-02",
                "date_to": "2016-11-03",
                "report_type": "order_report",
            }
        )
        self.assertTrue(form.is_valid())

    def test_date_range_incomplete(self) -> None:
        form = forms.ReportForm(
            data={
                "report_type": "order_report",
                "date_to": "",
                "date_from": "2016-11-02",
            }
        )
        self.assertTrue(form.is_valid(), form.errors)
        form = forms.ReportForm(
            data={
                "date_from": "",
                "date_to": "2016-11-03",
                "report_type": "order_report",
            }
        )
        self.assertTrue(form.is_valid())
        form = forms.ReportForm(
            data={"date_from": "2016-11-02", "report_type": "order_report"}
        )
        self.assertTrue(form.is_valid())
        form = forms.ReportForm(
            data={"date_to": "2016-11-03", "report_type": "order_report"}
        )
        self.assertTrue(form.is_valid())

    def test_date_range_incorrect(self) -> None:
        form = forms.ReportForm(
            data={
                "date_from": "2016-11-03",
                "date_to": "2016-11-02",
                "report_type": "order_report",
            }
        )
        self.assertFalse(form.is_valid())

    def test_date_range_equal(self) -> None:
        form = forms.ReportForm(
            data={
                "date_from": "2016-11-02 00:00:00",
                "date_to": "2016-11-02 00:00:00",
                "report_type": "order_report",
            }
        )
        self.assertFalse(form.is_valid())



from __future__ import annotations

from datetime import datetime
from typing import Any

from django import forms
from django.utils.translation import gettext_lazy as _
from oscar.forms.widgets import DateTimePickerInput

from .utils import GeneratorRepository


class ReportForm(forms.Form):
    generators = GeneratorRepository().get_report_generators()
    type_choices = []
    for generator in generators:
        type_choices.append((generator.code, generator.description))

    report_type = forms.ChoiceField(
        widget=forms.Select(),
        choices=type_choices,
        label=_("Report Type"),
        help_text=_("Only the offer and order reports use the selected date range"),
    )

    date_from = forms.DateTimeField(
        label=_("Date from"),
        required=False,
        widget=DateTimePickerInput,
    )

    date_to = forms.DateTimeField(
        label=_("Date to"),
        help_text=_("The report is inclusive of this date"),
        required=False,
        widget=DateTimePickerInput,
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # get any newly registered generators
        type_choices = [
            ((generator.code, generator.description)) for generator in self.generators
        ]
        self.fields["report_type"].choices = type_choices  # type:ignore[attr-defined]

    def clean(self) -> dict[str, str | datetime | None]:
        date_from = self.cleaned_data.get("date_from", None)
        date_to = self.cleaned_data.get("date_to", None)
        if (
            all([date_from, date_to])
            and self.cleaned_data["date_from"] >= self.cleaned_data["date_to"]
        ):
            raise forms.ValidationError(
                _("Your start date must be before your end date")
            )
        return self.cleaned_data
